package edu.washington.cac.myuw.servlet.session;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.servlet.http.HttpServletRequest;

import edu.washington.cac.myuw.util.SynchObject;
import edu.washington.cac.myuw.servlet.Log;
import edu.washington.cac.myuw.servlet.Memory;

/** Manages the pool of UWSessions, looking for memory and security problems

   @author Greg Barnes
   @version 1.1
 */
/*
   Limitations:  under Resin, if an HttpSession is invalidated, this affects
   sessions for all webapps.  So, for example, if you switch from a user
   to a guest in the same webapp, we invalidate the HttpSession, which logs
   you out of other webapps on the same machine.  Ideally, we would allow
   1 user and 1 guest session per webapp to avoid having to invalidate.
 */
/* Testing:
  The following tests should be run to insure this class and its related
  classes are working:

  Note that getting the log messages, particularly about the oldest user(s),
  can be made easier by changing the STATUS_INTERVAL variable to something
  lower.
   
  1) Login to an application and move around.  
     Expected result:  everything should work normally.
  2) Login to an application and log out.  Then go back and try to
     access pages without logging in again.  
     Expected result: logging out should work; attempts to access pages 
     that haven't been accessed before should fail with an invalid session 
     message.
  3) Login, then login again as the same user (don't log out).  
     Expected result: everything should appear fine to the user.  A 
     message should be printed to the log about reusing the old session.
  3a) Login as guest, then login again as guest (don't log out).  
     Expected result: everything should appear fine to the user.  A 
     message should be printed to the log about reusing the old session (and
     changing the new guest random number to the old).
  4) Login as a user, then login as guest.  
     Expected result:  everything should appear fine to the user.  A log
     message should appear about how the old session was invalidated.
  5) Login, kill the browser, then login as another user.
     Expected result:  everything should work fine.  Old user should still
     be in the SessionManager (this is easiest to tell if the old user is
     the 'Oldest' user).
  6) Login, kill the browser, then login as the same user.
     Expected result:  everything should work fine.  A log message should
     appear saying a foreign session with the same uid was invalidated.
  7) Login, kill the browser, and wait for the session to be invalidated
     by the application server.  Then login as a different user and check
     that the old session was invalidated.
     Expected result:  everything should appear fine to the user.  A log 
     message should appear saying the invalidated session was removed.
     Invalidations can always be temporarily forced by changing the code if 
     it takes too long to wait for the application server to invalidate.
  8) On a single browser, login twice as the same user at the same time
     (make sure the logins overlap somewhat).  It is best to do this on
     an application where the login procedure takes some time, such as MyUW.
     Expected result:  Depending on the timing of the logins, either both
     will work correctly, and the same log message as in #3 will appear, or
     you will get the same results as #10.  It depends on whether
     one gets the session cookie before the second asks for a session cookie;
     if this happens, the SessionManager can tell they are the same browser
     and reuse the session.  Otherwise, it must assume they are different
     browsers, and invalidates the older session.  Finally, if there are 
     multiple application servers, it's possible the two could be sent to 
     different servers, in which case nothing at all strange may appear in 
     the log (I have yet to have this happen, so I don't know what would 
     happen here).
  9) As #8, but login with the userid of a 'new' user, where the login
     process makes a substantial change to the database and should not be
     done twice.
     Expected result:  As #8, except the chances of the SessionManager seeing
     two 'different' browsers is greater.  In any event, under no 
     circumstances should the new user process be performed twice in the 
     database for the same user, although guaranteeing that is beyond the 
     scope of this code.
  10) Using two browsers, login as the same user at the same time.
     Expected result:  The 'first' browser will get an invalid session
     message; the second works fine.  Same log message as #6.
  11) Using two browsers, login as a user on one, then login as the
     same user on the other.  Then move around in the first browser's 
     application
     Expected result:  The first browser should get an invalid session
     message.  The log will have the same message as #6.
  12) Login as the oldest user (you may have to restart the application
     server).  Kill the browser, wait a few minutes, then login as another
     user.  
     Expected result:  The oldest user's age is roughly correct in the
     log message.
  13) Login as the oldest user.  Keep this browser active.  Using another
     browser, create approximately 8 different users (this is most easily
     accomplished by starting the browser, logging in as guest, then killing
     the browser).  Go back to the first browser and move around a bit.
     (don't log out).  Then kill the browser and login as a different
     user.
     Expected result:  This is all to test the orderedSessions map.  Moving
     around as the same user will force the 'oldest' session to become
     the newest, thus testing that the ordering is correct.  Nothing should
     appear out of the ordinary, but in the log, the oldest user will change
     from the oldest one on the first browser to the first of the 8
     users created on the second browser.  (This will happen when we
     move around in the first browser, but usually this information is
     not printed to the log until a new session is created, which is
     why we kill the browser and login as another user).
 */

/* Note on correctness:  The 3 entry methods to the class (loadOldSession,
   createNewSession, and invalidate) are synchronized on the class lock.
   This should guarantee the data structures are always
   viewed in a consistent state (as long as the code maintains that state),
   since there will be at most one thread running any of the methods at
   any time.
 */
/* Note on deadlock:
  3 locks are used below:

  The invalidateSynch lock prevents two threads from invalidating a session
  simultaneously.  This prevents a Resin bug, and should be removed when
  the Resin bug is fixed.  
  
  This lock cannot cause deadlock because once it is acquired, 
  no other locks need to be acquired before it is released.

  The class lock [synchronized static]
  is used to prevent two threads from changing the main data structures 
  simultaneously.  

  The lock on a UWSession object is used if a thread is operating
  on a UWSession to insure that no other thread will change the UWSession
  state in the interim.  Most of the time a thread is operating on its
  own session, but the
  SessionManager should be able to institute policies where one thread
  invalidates another thread's session.  This is where deadlock can occur.

  We will prevent deadlock by insuring that the class lock is always
  acquired before any UWSession locks.  This is done in the UWSession
  class, which does not hold any locks when SessionManager methods are
  called.

  We could relax this a bit by allowing locks when createNewSession is
  called and insuring a few other invariants.  This is because when
  a UWSession calls createNewSession, it's easy to guarantee the session
  isn't already in the SessionManager data structures.  Hence, no
  other thread can try to lock it because it cannot know about it, and
  deadlock loops are avoided.  But it is conceptually much easier just
  to create this hierarchy of locks:  class lock, UWSession lock, 
  invalidateSynch lock, and guarantee that if a thread acquires more than
  one of these locks, it always acquires the locks in order from first
  to last.
 */
public class SessionManager 
{
  /** Used to synch access to session invalidations */
  private static SynchObject invalidateSynch = new SynchObject();

  /* The data structures below allow us to retrieve an appropriate object
    of one of these classes (UWSession, HttpSessionAndId, userid) from an 
    object of the other two, as follows:  

    httpSessions.get(UWSession) returns the associated HttpSessionAndId.
    orderedSession.get(HttpSessionAndId.getId()) returns the associated userid.
    sessions.get(userid) returns an associated UWSession.

    Notes on HttpSessions:
      HttpSessions should not be used as keys in the data structures, as 
      a servlet server may not return the same HttpSession every time.  
      The HttpSession's id (HttpSession.getId()), however, should remain 
      the same.

      HttpSessions should not be used as values in the data structures, 
      either, since we need to get the id associated with the session
      to get the userid, and some servlet containers won't return an id
      for invalidated sessions.  We use HttpSessionAndId objects instead.
      [In fact, below and in SessionComparator, we don't
      use HttpSession objects at all, always accessing HttpSessions from
      an HttpSessionAndId object.]
    
      UWSession.getSession returns an HttpSession, but this
      session may be different from what is in our data structures, so
      we use the httpSessions map to associate UWSessions with HttpSessions

      There may be more than one UWSession associated with an 
      HttpSession (and thus, a userid).  
      
    THESE DATA STRUCTURES ARE NOT MEANT TO CORRESPOND EXACTLY WITH THE 
    STATE OF THE APPLICATION MANAGER'S SESSION MANAGER.  They are merely 
    meant to approximate this state.  What is most important is that the 
    data structures be kept *internally consistent*.  If at any time we 
    notice a discrepancy between the SessionManager's state and the 
    application manager's state, we will try to correct it (usually be 
    invalidating the sessions and removing them from the data structure).
    But at no time (except when we are changing them) should there
    be an inconsistency in the SessionManager's data structures themselves.

    Note also that SessionComparator keeps track of the last accessed times
    of all the HttpSessions in its own map (see the comments in that class
    to learn why the HttpSession.lastAccessedTime method is not useful for
    this).  So whenever a session is added, changed, or removed from our
    data structures, SessionComparator's map must also be updated.
   */

  /** Keeps track of all UWSessions, by userid.  
      Key = userid, Value = UWSession 
    */
  private static Map sessions = new HashMap();

  /** Keeps track of all HttpSessions, by UWSession.
      Key = UWSession, Value = HttpSessionAndId
    */
  private static Map httpSessions = new HashMap();

  /** Keeps track of all guest HttpSessions, ordered by last modified time.
    Key = HttpSession's id, value = associated 'userid' (for guests, a 
    hopefully unique random number)
   */
  private static SortedMap guestSessions = 
      new TreeMap(new SessionComparator());

  /** Keeps track of all HttpSessions, ordered by last modified time.
    Key = HttpSession's id, value = associated userid
   */
  private static SortedMap orderedSessions = 
      new TreeMap(new SessionComparator());

  /** A file to print messages to */
  private static Log log = new Log();

  /** A count of the number of sessions created so far */
  private static long totalSessions = 0;

  /** # of new sessions created before we print the SessionManager status
    i.e., if the interval is 10, the status is printed after 10, 20, 30, etc.
    creates
   */
  private static final long STATUS_INTERVAL = 10L;

  /** Desired limit to number of megabytes used by the application server.
      If the actual used memory goes above this limit, we may invalidate
      old sessions
   */
  private static final long MEMORY_USAGE_LIMIT = 500L;

  /** Desired minimum age (in minutes) below which we will not delete
      an 'old' session.
   */
  private static final long DELETE_LIMIT = 15L;

  /** Desired minimum age (in minutes) below which we will not delete
      an 'old' guest session.
   */
  private static final long GUEST_DELETE_LIMIT = 10L;

  /** Number of old sessions to delete before creating a new one when
      memory is low
   */
  private static final int OLD_FOR_NEW = 2;

  /** Factor by which to multiply the age of guest sessions when comparing
      them to user sessions for the purpose of deleting them.  The higher
      the factor, the less important a guest session.  E.g., if 
      GUEST_FACTOR = 2, we will delete a guest session before we delete a
      user session less than twice its age.  Note, however, that the 
      GUEST_DELETE_LIMIT is absolute:  e.g., if the GUEST_DELETE_LIMIT and 
      the DELETE_LIMIT are both 15 minutes and the GUEST_FACTOR is 5, a guest 
      session 10 minutes old will not be deleted, even if the oldest 
      session is less than 50 minutes old.
   */
  private static final float GUEST_FACTOR = 2.0F;

  /**
    @return Whether a session exists in the data structures
    @param session The session in question
   */
  private static boolean existsInDataStructures(HttpSessionAndId session) 
  {
    return session != null && orderedSessions.containsKey(session.getId());
  }

  /**
    @return Whether a session exists in the data structures
    @param session The session in question
   */
  private static boolean existsInDataStructures(UWSession session) 
  {
    return session != null && httpSessions.containsKey(session);
  }

  /**
    @return Whether a session for the user exists in the data structures
    @param userid The user in question.  For guests, the integer id of the
      guest must be passed in, not UWSession.GUEST_ID.
   */
  private static boolean existsInDataStructures(String userid) 
  {
    return userid != null && sessions.containsKey(userid);
  }

  /** 
    Does a set of parameters exist in the data structures and
       do they correspond to the same session?
    @param uwSession UWSession to check
    @param httpSession HttpSession to check
    @param userid userid to check
    @return Whether a set of parameters are in the data structures and
       correspond to the same session
   */
  private static boolean existsInDataStructures(UWSession uwSession,
                                                HttpSessionAndId httpSession,
                                                String userid)
  {
    return (existsInDataStructures(uwSession) && 
            existsInDataStructures(httpSession) &&
            existsInDataStructures(userid) &&
            uid(httpSession).equals(userid) &&
            ((HttpSessionAndId) httpSessions.get(uwSession)).getId() == 
                httpSession.getId() &&
            sessions.get(userid) == uwSession);
  }

  /** 
    Given a set of parameters that do not exist in the data structures,
    what is wrong with them?
    @param uwSession UWSession to check
    @param httpSession HttpSession to check
    @param userid userid to check
    @return A printable set of reasons why a call to existsInDataStructures
    (above) returns false.
   */
  private static String whyNotExistsInDataStructures(UWSession uwSession,
      HttpSessionAndId httpSession, 
      String userid)
  {
    String reason = "";

    if (!existsInDataStructures(uwSession)) {
      reason += "UWSession is not in the data structures.  ";
    } else if (((HttpSessionAndId) httpSessions.get(uwSession)).getId() != 
               httpSession.getId())
    {
      reason += "uwSession's httpSession " + 
                ((HttpSessionAndId) httpSessions.get(uwSession)).getId() +
                " does not match the httpSession " + 
                httpSession.getId() + ".  ";
    }


    if (!existsInDataStructures(httpSession)) {
      reason += "HttpSession is not in the data structures.  ";
    }

    if (!existsInDataStructures(userid)) {
      reason += "userid is not in the data structures.  ";
    }

    if (!userid.equals(uid(httpSession))) {
      reason += "HttpSession's uid " + uid(httpSession) + 
                " does not match the userid.  ";
    }

    if (sessions.get(userid) != uwSession) {
      reason += "userid's uwSession " + sessions.get(userid) + 
                " does not match the uwSession.";
    }

    if (reason == "") {
      return "It DOES exist in the data structures!";
    } else {
      return reason;
    }
  }

  /** 
    Do an httpSession and userid exist in the data structures
      and correspond to the same session?
    @param httpSession HttpSession to check
    @param userid userid to check
    @return Whether the httpSession and userid are in the data structures
      and correspond to the same session
   */
  private static boolean existsInDataStructures(HttpSessionAndId httpSession,
                                                String userid)
  {
    return (existsInDataStructures(httpSession) &&
            existsInDataStructures(userid) &&
            uid(httpSession).equals(userid));
  }

  /**
    @return The oldest guest session, or null if there are no guest sessions
   */
  private static HttpSessionAndId oldestGuest() 
  {
    return guestSessions.size() > 0 ? 
        (HttpSessionAndId) httpSessions.get(
            (UWSession) sessions.get(
                uid((String) guestSessions.firstKey()))) : null;
  }

  /**
    @return The oldest session, or null if there are no sessions
   */
  private static HttpSessionAndId oldest() 
  {
    return orderedSessions.size() > 0 ? 
        (HttpSessionAndId) httpSessions.get(
            (UWSession) sessions.get(
                uid((String) orderedSessions.firstKey()))) : null;
  }

  /**
    @return The name of the user represented by a session in our data 
      structures
    @param session The id of the session in question
   */
  private static String uid(String id) 
  {
    return (String) orderedSessions.get(id);
  }

  /**
    @return The name of the user represented by a session in our data 
      structures
    @param session The session in question
   */
  private static String uid(HttpSessionAndId session) 
  {
    return uid(session.getId());
  }

  /** The number of milliseconds in a minute */
  private static int MILLISECONDS_PER_MINUTE = 1000 * 60;

  /**
    The current time, in milliseconds since the epoch started
    @return The current time, in milliseconds since the epoch started
   */
  private static long currentTime()
  {
    return new Date().getTime();
  }

  /**
    @return the age of a session in minutes
    @param session The session in question
   */
  private static long age(HttpSessionAndId session) 
  {
    return (currentTime() -
            SessionComparator.getLastAccess(session.getId()).longValue())
           / MILLISECONDS_PER_MINUTE;
  }

  /**
    Print out particulars on an <code>HttpSession</code> in our data structures
    @param log Log to print message to
    @param session Session to print out
    @param prefix Prefix to append to the message
   */
  private static void printSession(Log log, 
                                   HttpSessionAndId session, 
                                   String prefix)
  {
    log.println(prefix + uid(session) + ": " + age(session) + " minutes old.");
  }

  /**
    print out the status of the data structures 
    @param log Log to print messages to
   */
  private static void printStatus(Log log) 
  {
    log.println("SessionManager::printStatus: " + sessions.size() + " " + 
                httpSessions.size() + " " +
                guestSessions.size() + " " + orderedSessions.size() + " " +
                totalSessions);
    new Memory().printStatus(log);

    if (sessions.size() > 0) {
      printSession(log, oldest(), "Oldest session ");
    }

    if (guestSessions.size() > 0) {
      printSession(log, oldestGuest(), "Oldest guest session ");
    }
  }
    
  /**
    Check the data structures, and maybe print out some things
   */
  private static void checkStatus() 
  {
    if (sessions.size() != orderedSessions.size() || 
        httpSessions.size() != orderedSessions.size() || 
        guestSessions.size() > sessions.size()) 
    {
      log.println("Data structures in inconsistent state!");
    }
  }

  /**
    Update the last access time of a session in the data structures
    @param session Session to update
    @exception IllegalArgumentException if the session is null, or not in
      the data structures
   */
  private static void updateLastAccess(HttpSessionAndId session) 
      throws IllegalArgumentException
  {
    if (!existsInDataStructures(session)) {
      throw new IllegalArgumentException(
          "SessionManager::updateLastAccess: " + session);
    }

    /* Last accessed time is automatically updated in the sessions 
       themselves, but the SortedMaps must go by the last access time
       kept by the SessionComparator.  To properly update the time, we
       must remove the session from the SortedMaps, update the 
       last access time in SessionComparator, then reinsert the session 
       into the SortedMap(s).
     */
    // retrieve info we'll need later
    String uid = (String) orderedSessions.get(session.getId());
    boolean isGuest = guestSessions.containsKey(session.getId());

    // remove from SortedMaps
    orderedSessions.remove(session.getId());
    guestSessions.remove(session.getId());

    // update last access time
    SessionComparator.setLastAccess(session);

    // put back in SortedMaps
    orderedSessions.put(session.getId(), uid);
      
    if (isGuest) {
      guestSessions.put(session.getId(), uid);
    }
  }

  /**
    Remove a session from all the data structures
    @param uwSession UWSession to remove
    @param httpSession HttpSession to remove
    @param userid userid to remove
    @exception IllegalArgumentException If one of the parameters is null
      or not in the data structures, or if the parameters do not correspond
      to the same session
   */
  private static void removeSession(UWSession uwSession,
                                    HttpSessionAndId httpSession,
                                    String userid) 
      throws IllegalArgumentException
  {
    if (!existsInDataStructures(uwSession, httpSession, userid)) {
      throw new IllegalArgumentException("SessionManager::removeSession: " +
          uwSession + " " + httpSession.getId() + " " + userid + ":" + 
          whyNotExistsInDataStructures(uwSession, httpSession, userid));
    }

    sessions.remove(userid);
    httpSessions.remove(uwSession);
    orderedSessions.remove(httpSession.getId());
    guestSessions.remove(httpSession.getId());
    SessionComparator.remove(httpSession);
  }

  /**
    Remove a <code>UWSession</code> from all the data structures
    @param session UWSession to remove
   */
  private static void removeSession(UWSession session) 
  {
    removeSession(session, (HttpSessionAndId) httpSessions.get(session), 
        (String) orderedSessions.get(
            ((HttpSessionAndId) httpSessions.get(session)).getId()));
  }

  /**
    Remove an HttpSession from the data structures
    @param session Session to remove
   */
  private static void removeSession(HttpSessionAndId session) 
  {
    removeSession(
        (UWSession) sessions.get(orderedSessions.get(session.getId())), 
        session, (String) orderedSessions.get(session.getId()));
  }

  /**
    Remove a userid from the data structures
    @param userid userid to remove
   */
  private static void removeSession(String userid) 
  {
    removeSession((UWSession) sessions.get(userid),
                  (HttpSessionAndId) httpSessions.get(sessions.get(userid)),
                  userid);
  }

  /**
    Add a session to all the data structures
    @param uwSession UWSession to add
    @param httpSession HttpSession to add
    @param userid unique userid
    @param log File to print messages to
    @exception IllegalArgumentException If any of the arguments are null,
      or they exist in the data structures already
   */
  private static void addSession(UWSession uwSession, 
                                 HttpSessionAndId httpSession,
                                 String userid,
                                 Log log) 
  {
    if (uwSession == null || httpSession == null || 
        httpSession.getSession() == null || userid == null) 
    {
      throw new IllegalArgumentException(
          "SessionManager::addSession: null argument: " + uwSession + " " +
          httpSession + " " + httpSession.getSession() + " " + userid);
    }

    if (existsInDataStructures(uwSession) || existsInDataStructures(userid) ||
        existsInDataStructures(httpSession))
    {
      throw new IllegalArgumentException(
          "SessionManager::addSession: session already exists: " +
          uwSession + " " + httpSession + " " + userid);
    }

log.println("addSession: " + uwSession + " " + httpSession.getId() + " " + userid);

    SessionComparator.setLastAccess(httpSession);
    sessions.put(userid, uwSession);
    httpSessions.put(uwSession, httpSession);
    orderedSessions.put(httpSession.getId(), userid);
  
    if (uwSession.isGuest()) {
      guestSessions.put(httpSession.getId(), userid);
    }
  }

  /*
    Resin sometimes doesn't delete sessions that are too old, which
    can cause problems.  So we need to be able to test a session's age
    ourselves.
   */
  /**
    Should a session already be invalidated?
    @param session Session to test
    @return Should the session already be invalidated?
   */
  private static boolean isTooOld(HttpSessionAndId session)
  {
    return currentTime() > session.getSession().getLastAccessedTime() + 
        session.getSession().getMaxInactiveInterval() * 1000L;
  }

  /**
    @return Is the session invalid?
    @param session Session to test
   */
  private static boolean invalid(HttpSessionAndId session) 
  {
    /* Sessions don't seem to come with an isInvalid method, so
       try to get an attribute, which is supposed to raise an
       exception if it's invalid
     */
    try {
      session.getSession().getAttribute("dummy");

      /* Since servlet containers might drop the ball on invalidating
         sessions, test the age ourselves
       */
      if (isTooOld(session)) {
log.println("Invalidating session " + session + "; age " + age(session) + " greater than max age " + session.getSession().getMaxInactiveInterval()/60);
        invalidateSession(session);
        return true;
      } else {
        return false;
      }
    } catch (IllegalStateException e) {
      return true;
    }
  }

  /**
    Remove sessions that have been invalidated by the servlet server
    from our data structures.
    @param log File to print messages to
   */
  private static void removeInvalidSessions(Log log) 
  {
    /* Invalidated sessions should have been invalidated due to their
       age, so just start with the oldest and keep removing until
       we find a session that hasn't been invalidated 
     */
    for (HttpSessionAndId s = oldest(); s != null && invalid(s); s = oldest())
    {
      /* Note that there is no need to synchronize on the UWSession,
         as the invalidate method does, as that is meant to
         prevent someone who is using the session from having it invalidated
         out from under them.  But this session has already been invalidated,
         so there is no reason to be concerned about that.
       */
log.println("Removing invalidated session for user " + uid(s));
      removeSession(s);
    }
  }

  /**
    @return Is the available memory low?
   */
  private static boolean memoryLow() 
  {
    return new Memory().getUsedMemory() > MEMORY_USAGE_LIMIT;
  }

  /**
    Invalidate an <code>HttpSession</code>, remove it from our data structures,
    and try to invalidate the associated <code>UWSession</code>

    Note that many UWSessions may hold the same HttpSession, therefore
    a particular UWSession may find its HttpSession invalidated out from 
    under it without warning after this method is run.  However, the 
    assumption here is that the session is old and probably abandoned.

    @param session HttpSession to invalidate
   */
  private static void invalidateAndRemove(HttpSessionAndId session) 
  {
    invalidate((UWSession) sessions.get(orderedSessions.get(session.getId())));
  }

  /**
    Invalidate the oldest session and remove it from the data structures
    @param log File to print messages to
   */
  private static void removeOldestUser(Log log) 
  {
    printSession(log, oldest(), 
        "SessionManager::removeOldSession: removing oldest session ");
    invalidateAndRemove(oldest());
  }

  /**
    Invalidate the oldest guest session and remove it from the data structures
    @param log File to print messages to
   */
  private static void removeOldestGuest(Log log) 
  {
    printSession(log, oldestGuest(), 
        "SessionManager::removeOldSession: removing oldest guest session ");
    invalidateAndRemove(oldestGuest());
  }

  /**
    @return Is there a user whose session can be deleted according to
      GUEST_DELETE_LIMIT?
   */
  private static boolean canDeleteGuest() 
  {
    return oldestGuest() != null && age(oldestGuest()) >= GUEST_DELETE_LIMIT;
  }

  /**
    @return Is there a user whose session can be deleted according to
      DELETE_LIMIT?
   */
  private static boolean canDeleteUser() 
  {
    return oldest() != null && age(oldest()) >= DELETE_LIMIT;
  }

  /**
    Try to remove an old session (presumably because memory is too low)
    @param log File to print messages to
   */
  private static void removeOldSession(Log log) 
  {
//printStatus(log);

    if (!canDeleteGuest() && !canDeleteUser()) {
      log.println(
          "SessionManager::removeOldSession: no sessions old enough to delete");
    } else if (!canDeleteGuest()) {
      removeOldestUser(log);
    } else if (!canDeleteUser()) {
      removeOldestGuest(log);
    } else {
      if (age(oldestGuest()) * GUEST_FACTOR >= age(oldest())) {
        removeOldestGuest(log);
      } else {
        removeOldestUser(log);
      }
    }
  }

  /**
    Perform necessary tasks before creating a new session
    @param log File to print messages to
   */
  private static void preCreate(Log log) 
  {
    try {
      removeInvalidSessions(log);
    } catch (RuntimeException e) {
    // don't stop people from logging in because of this
      e.printStackTrace();
    }

    try {
      if (memoryLow()) {
        for (int i = 0; i < OLD_FOR_NEW; i++) {
          removeOldSession(log);
        }
      }
    } catch (RuntimeException e) {
    // don't stop people from logging in because of this
      e.printStackTrace();
    }
  }

  /**
    Perform necessary tasks after creating a new session
    @param log File to print messages to
   */
  private static void postCreate(Log log) 
  {
    totalSessions++;
    checkStatus();

    if ((totalSessions % STATUS_INTERVAL) == 0) {
      printStatus(log);
      System.gc();
    }
  }

  /**
    Invalidate any old sessions corresponding to the new one we will create
    @param session <code>UWSession</code> for which we are creating a new
      <code>HttpSession</code>
    @param uid Userid for whom we're creating a session
    @param log File to print messages to
   */
  private static void invalidateOld(UWSession session, String uid, Log log)
  {
    if (existsInDataStructures(uid)) {
      log.println("SessionManager::createNewSession - Invalidating foreign session with same uid");
      invalidate((UWSession) sessions.get(uid));
    }

    if (existsInDataStructures(session)) {
      log.println("SessionManager::createNewSession - session already in data structures; invalidating");
      invalidate(session);
    }
  }

  /**
    Ensure an <code>HttpSession</code> we are about to add to our data
    structures isn't there already

    @param session HttpSession in question
    @param log File to print messages to
   */
  private static void removeOld(HttpSessionAndId session, Log log)
  {
    if (existsInDataStructures(session)) {
      log.println("SessionManager::createNewSession - 'new' session already in data structures, will remove! " + session);
      // We need the session, so don't invalidate it, just remove it
      removeSession(session);
    }
  }

  /** @return does a <code>UWSession</code> contain an <code>HttpSession</code>
    @param session UWSession in question
   */
  private static boolean hasOldSession(UWSession session) 
  {
    return session.getSession() != null;
  }

  /**
    Can we use the existing <code>HttpSession</code> instead of
      creating a new one for a user
    @param session <code>UWSession</code> that may hold a useful old
      HttpSession
    @param newUserid The new user's id
    @param isNewGuest Is the new user a guest?
    @return Can we use the existing <code>HttpSession</code> instead of
      creating a new one for a user
   */
  private static boolean canUseOldSession(UWSession session, 
                                          String newUserid,
                                          boolean isNewGuest)
  {
    /* can reuse if the old session exists, matches the new userid, 
       and is in data structures
     */
    if (isNewGuest) {
      return hasOldSession(session) && 
             session.isGuest() && 
             guestSessions.containsKey(session.getSession().getId());
    } else {
      return hasOldSession(session) && 
             session.represents(newUserid) &&
             existsInDataStructures(new HttpSessionAndId(session.getSession()), 
                                    newUserid);
    }
  }
    
  private static final String USING_OLD_SESSION = 
      "SessionManager::createNewSession - Using old session";

  /**
    Create a new <code>HttpSession</code> for a <code>UWSession</code>
    @param req The HTTP request object
    @param session The <code>UWSession</code>
    @param uid Userid of the new session (for guests, pass in the unique
      integer id, not UWSession.GUEST_ID)
    @param isGuest Is the new session a guest session?
    @param log File to print messages to
   */
  static synchronized void createNewSession(HttpServletRequest req, 
                                            UWSession session, 
                                            String uid,
                                            boolean isGuest,
                                            Log log) 
  {
    preCreate(log);

    synchronized(session) {
      // Step 1:  Deal with any old HttpSessions

      if (canUseOldSession(session, uid, isGuest)) {
        if (isGuest) {
          log.println(USING_OLD_SESSION + " with id " + 
                      uid(new HttpSessionAndId(session.getSession())));
          // change the new session's log uid to the old session's
          log.setUid(uid(new HttpSessionAndId(session.getSession())));
        } else {
          log.println(USING_OLD_SESSION);
        }

        updateLastAccess(new HttpSessionAndId(session.getSession()));
        return;
      } 

      if (hasOldSession(session)) {
        log.println(
            "SessionManager::createNewSession - Invalidating old session");
        invalidate(session);
      }

      // Step 2: Insure that none of the parameters in the addSession call
      // below are already in the data structures (handled by invalidateOld and 
      // removeOld below).

      // Step 3: Create the new session and add it to our data structures
      // (req.getSession and addSession below).  Note that req.getSession
      // must be done before Step 2 can be completed

      invalidateOld(session, uid, log);

      /* this should get a new HttpSession, as any old one would have been
         invalidated in Step 1 */
      session.setSession(req.getSession(true)); 

      /* A new HttpSession shouldn't be in the data structures, 
         but you never know about application servers, so check anyway. */
      removeOld(new HttpSessionAndId(session.getSession()), log);

      addSession(session, new HttpSessionAndId(session.getSession()), uid, log);
    }

    postCreate(log);
  }

  /**
    Load the old <code>HttpSession</code> associated with this request
    into a <code>UWSession</code> object.  If there is no old session, do
    nothing.

    @param req The HTTP request object
    @param session The <code>UWSession</code>
    @exception IllegalStateException if the session is not new, but is
      not known to the manager
   */
  static synchronized void loadOldSession(HttpServletRequest req, 
                                          UWSession session)
  {
    synchronized(session) {
      session.setSession(req.getSession(false));

      if (session.getSession() == null) {
        // not a big deal
        return;
      } else if (existsInDataStructures(
          new HttpSessionAndId(session.getSession()), 
          session.userKey())) 
      {
        updateLastAccess(new HttpSessionAndId(session.getSession()));
      } else { 
        // can't use old session; have to invalidate it
        log.println(
            "SessionManager::loadOldSession: Invalidating unknown old session: "
            + session + " " + session.getSession().getId() + " " + 
            session.userKey());
        invalidate(session);
      }
    }

    checkStatus();
  }

  /** Invalidate an HttpSession
    @param session Session to invalidate
   */
  private static void invalidateSession(HttpSessionAndId session) 
  {
    /* could use the invalid() method above to try to test if session is
       invalid before calling HttpSession.invalidate(), but 
       it could always get invalidated between when we call "invalid()"
       and when we try to invalidate, and raise the exception anyway
     */
    try {
      /* Resin appears to have a bug that may appear if two threads try
         to invalidate the same session simultaneously.  So synchronize
         the access to avoid this 
       */     
      synchronized(invalidateSynch) {
        session.getSession().invalidate();
      }
    } catch (IllegalStateException e) {
      log.println("SessionManager::invalidateSession: session has already been invalidated.");
    }
  }

  /**
    Invalidate a <code>UWSession</code>

    @param session The session to invalidate
   */
  static synchronized void invalidate(UWSession session) 
  {
    /* Synchronize on the session so UWSession methods don't get an
       inconsistent view.  See UWSession.  
     */
    synchronized(session) {
      // Take the underlying HttpSession out of our data structures.

      if (existsInDataStructures(session)) {
        removeSession(session);
      }

      if (session.getSession() != null) {
        HttpSessionAndId hsai = new HttpSessionAndId(session.getSession());

        if (existsInDataStructures(hsai)) {
          removeSession(hsai);
        }

        invalidateSession(hsai);
        session.setSession(null);
      } else {
        log.println("SessionManager::invalidate: already invalidated: " + 
                    session);
      }
    }

    checkStatus();
  }
}
