package edu.washington.cac.myuw.test;

import java.io.*;

/** Generic stress tester.  Typically, a list of users is supplied in a file,
    and a number of tests are run using data associated with these users.
    For example, getting all relevant records for the given user.

  @author Greg Barnes
  @version 1.0 
*/
public abstract class StressTest extends Thread {
    /** the name of the file that contains info for the 'users' for the test */
    private String idsFile;

    /** the desired number of distinct users, and hence, the minimum number
        of users in the <code>idsFile</code>
      */
    private static final int DESIRED_SIZE = 181;	// a convenient prime

    /** the uwnetids of the users */
    protected static String[] uwnetids = new String[DESIRED_SIZE];

    /* Strategy:  we run a number of thread simultaneously, making requests
	associated with users in the uwnetids array.

	We want to avoid caching by a foreign database or other server, so
	the threads should not all be working on the same data at the same 
	time.  So we start them at random places in the array.

        Also, in case two threads get in sync, we don't want two threads to 
	issue the same series of requests (e.g., if two are working on
	requests for user 1, they should not both follow with requests for
	user 2).  To avoid this, we have them 'step' through the array at
	different intervals.  If thread A is working on user 1, it nexts
	works on user 2.  But thread B might work on every other user
	(1, 3, 5, ...), while thread C gets every 3rd user.

	The interval is the size of the gap between subsequent users.  To
	make sure each thread eventually makes requests for all users,
	the intervals and the size of the array(s) must be relatively
	prime.  To make doubly sure that two threads don't follow each
	others orders too closely, we make the intervals all relatively
 	prime as well
      */

    /** the interval between successive users for this thread */
    private int interval;

    /** set the name of the <code>idsFile</code>
	@param idsFile the name of the idsFiles
      */
    protected void setIdsFile(String idsFile) {
	this.idsFile = idsFile;
    }

    /** get the interval variable 
	@returns the interval
      */
    protected int getInterval() {
	return this.interval;
    }

    /** set the interval variable 
	@param interval the interval
      */
    private void setInterval(int interval) {
	this.interval = interval;
    }

    /** run a test on a user
	@param index the index of the user in the <code>uwnetids</code> array
	@exception Exception almost anything
      */
    protected abstract void runTest(int index) throws Exception;

    /** read an entry in the <code>idsFile</code>.  By default, just read
	a single line and store it in the <code>uwnetids</code> array, but 
	this can be overridden if we need to supply more complex info
	@param i which entry are we reading
	@param line the first line in the entry 
      */
    protected void readEntry(int i, String line) {
	uwnetids[i] = line;
    }

    /** read the information about the users and store it
      */
    private void readInfo() {
	BufferedReader in = null;

        try {
	    in = new BufferedReader(new FileReader(this.idsFile));
    	    String aline;
	    int i = 0;

	    while ((aline = in.readLine()) != null && i < DESIRED_SIZE) {
		readEntry(i, aline);
		i++;
	    }

	    if (i != DESIRED_SIZE) {
		System.err.println("Not enough netids: " + i);
	        System.exit(1);
	    }
        } catch (IOException e ) {
	    System.err.println(this.idsFile + " can't be read!");
	    System.exit(1);
	} catch (Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	} finally {
	    try {
	        in.close();
	    } catch (IOException e) {}
	}
    }

    /** run the tests for all users */
    public void run() {
        long startTime = System.currentTimeMillis();
        int i;

        // start at a random position in the array (see Strategy comment above)
	int index = (int) (java.lang.Math.random() * DESIRED_SIZE);
	System.out.println(this.interval + ": " + index);

 	for (i = 0; i < DESIRED_SIZE; i++) {
	    try {
	        runTest(index);
            } catch (Throwable e) {
  		e.printStackTrace();
	    }

  	    index = (index + this.interval) % DESIRED_SIZE;
 	}

	long totalTime = System.currentTimeMillis() - startTime;
	System.out.println( this.interval + " ending. Time= " + totalTime + 
	      " milliseconds.");
    }

    /** prime numbers up to a certain point */
    private static final int[] PRIMES = {1, 2, 3, 5, 7, 11, 13, 17, 19, 23, 29,
      31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79};
    /** The default number of threads.  */
    private static final int NUMBER_OF_THREADS = 5;

    /** Run some stress tests.
	@param args Usage: StressTest <class> [<numberOfThreads>]
	  First argument is the name of a subclass of this class (the
	  test to run).
	  The optional second argument is the number of threads to run in
	  the test.  If not supplied, a default will be used.
	@exception ClassNotFoundException if the given class doesn't exist
	@exception InstantiationException if the given class can't be created
	@exception IllegalAccessException if the given class can't be created
      */
    public static void main(String[] args) 
      throws ClassNotFoundException, InstantiationException, 
	IllegalAccessException
    {
	int i, numberOfThreads = NUMBER_OF_THREADS;
	
	if (args.length > 1) {
	    numberOfThreads = new Integer(args[1]).intValue();
	}

System.err.println(numberOfThreads);
	StressTest st = (StressTest) Class.forName(args[0]).newInstance();
	st.readInfo();

	for (i = 0; i < numberOfThreads; i++) {
	    st = (StressTest) Class.forName(args[0]).newInstance();

	    // see Strategy comment above.  
	    st.setInterval(PRIMES[i]);  
	    st.start();
	}
    }
}
