// java

package uwcal;

import java.io.IOException;
import java.sql.SQLException;

import javax.servlet.http.HttpServletResponse;

import org.apache.ecs.Entities;

import edu.washington.cac.calendar.CalParameter;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.Cache;
import edu.washington.cac.calendar.data.CacheI;
import edu.washington.cac.calendar.data.CaldataException;
import edu.washington.cac.calendar.data.CaldataWarning;
import edu.washington.cac.calendar.data.CalendarObject;
import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.EventsI;
import edu.washington.cac.calendar.data.ItemException;
import edu.washington.cac.calendar.data.ItemAccessException;
import edu.washington.cac.calendar.data.Keyword;
import edu.washington.cac.calendar.data.Location;
import edu.washington.cac.calendar.data.Sponsor;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.calendar.db.Caldata;
import edu.washington.cac.myuw.servlet.Log;
import edu.washington.cac.myuw.servlet.session.UWSession;

import uwcal.out.Page;
import uwcal.out.DisplayText;

/**
  An action in the calendar application that acts on the database

  @author Greg Barnes
  @version 2.0
 */
public class DBAction extends CalAction 
{
  /** The message to be displayed if the action succeeds */
  private String successMessage;

  /** Holds any warnings generated for later display */
  private CaldataWarning w = null;        

  /** The object the action is performed on */
  private CalendarObject c = null;  

  /** The Cache where you can find <code>c</code> */
  private CacheI cache = null;           

  /**
    Create a minimal DBAction, not suitable for performing actions.
    This object can be used to match function names with the proper 
    DBAction class, and calling newInstance to create a more complete 
    version of the object which can perform the action

    @param function Name of the function this DBAction can handle
    @param successMessage message to be displayed if the action succeeds
   */
  DBAction(String function, String successMessage) 
  {
    super(function);
    this.successMessage = successMessage;
  }

  /** Create a new DBAction to perform actions

    @param function The function that was called
    @param cal The day for the action
    @param user The user performing the action
    @param es The EventsI object for the user
    @param cp The form parameters
    @param log A file to print messages to
    @param successMessage Message to display if the action succeeds
   */
  protected DBAction(String function,       
                     MyCalendar cal,
                     User user, 
                     EventsI es,
                     CalParameter cp,
                     Log log,
                     String successMessage)
  {
    super(function, cal, user, es, cp, log);
    this.successMessage = successMessage;
  }

  /** Create a new DBAction
    @param cal The day for the action
    @param user The user performing the action
    @param es The EventsI object for the user
    @param cp The form parameters
    @param log A file to print messages to
   */
  CalAction newInstance(MyCalendar cal, 
                        User user, 
                        EventsI es, 
                        CalParameter cp,
                        Log log) 
  {
    return new DBAction(getHandlesFunction(), cal, user, es, cp, log, 
                        this.successMessage);
  }

  /**
    Set the message to display if the action succeeds
    @param successMessage The message to display if the action succeeds
   */
  protected void setSuccessMessage(String successMessage) 
  {
    this.successMessage = successMessage;
  }

  /**
    Get the message to display if the action succeeds
    @return The message to display if the action succeeds
   */
  protected String getSuccessMessage() 
  {
    return this.successMessage;
  }

  /**
    @return The Cache that contains the object to operate on
   */
  protected CacheI getCache() 
  {
    return this.cache;
  }

  /**
    Create the main Cache object associated with this DBAction
   */
  protected void setCache() 
  {
    if (getHandlesFunction().endsWith(CalParameter.KEYWORD)) {
      this.cache = es.getKeywords();
    } else if (getHandlesFunction().endsWith(CalParameter.SPONSOR)) {
      this.cache = es.getSponsors();
    } else if (getHandlesFunction().endsWith(CalParameter.LOCATION)) {
      this.cache = es.getLocations();
    } else if (getHandlesFunction().endsWith(CalParameter.EVENT) || 
      getHandlesFunction().endsWith(CalParameter.FULLEVENT) ||
      getHandlesFunction().endsWith(CalParameter.EVENTREF)) 
    {
      this.cache = es;
    } else {
      throw new IllegalStateException("setCache: illegalAction " + 
                                      getHandlesFunction());
    }
  }

  /**
    Create the main object associated with the DB action.

    @exception CaldataException If there's a problem with the data
      entered.
    @exception SQLException An action might, as a side effect, also
      create another object.  This exception may be raised when trying
      to add the other object to the database.
    @exception ItemAccessException If the action tries to create another
      item as a side effect, and the user does not have permission to do
      so
   */
  protected void setCalendarObject() 
      throws CaldataException, SQLException, ItemAccessException
  {
    if (getHandlesFunction().endsWith(CalParameter.KEYWORD)) {
      this.c = new Keyword(cp, this.user);
    } else if (getHandlesFunction().endsWith(CalParameter.SPONSOR)) {
      this.c = new Sponsor(cp, this.user);
    } else if (getHandlesFunction().endsWith(CalParameter.LOCATION)) {
      this.c = new Location(cp, this.user);
    } else if (getHandlesFunction().endsWith(CalParameter.EVENT) 
               || getHandlesFunction().endsWith(CalParameter.FULLEVENT)) 
    {
      Event e = new Event(cp, this.user);
      w = e.getWarnings();

      // add/edit events can add new locations
      Location l;

      try {
        l = new Location(cp, this.user);
        /* If the user enters a link for an event, the above call
           will assign the link to the location as well.  To make
           sure this doesn't happen, change the location's link
           to null 
         */
        l.setLink(null);       
      } catch (CaldataException ex) {        // new location not entered
        l = null;
      }

      if (l != null) {
        // new location was entered
        es.getLocations().add(l);
        e.setLocation(l);
        message = DisplayText.ADDED_LOCATION_MSG;
      }

      this.c = e;
    } else {
      throw new IllegalStateException("setCalendarObject: illegalAction " + 
        getHandlesFunction());
    }
  }

  /** Set up the variables needed to perform the action
    @exception CaldataException If there's a problem with the form data
    @exception SQLException If there's a problem with the database
    @exception ItemAccessException If the action tries to do something to
      an item that it shouldn't
   */
  protected void setVars() 
      throws CaldataException, SQLException, ItemAccessException
  {
    setCache();
    setCalendarObject();
  }
    
  /**
    Perform the portion of the action that accesses the database

    @exception SQLException If there's a problem with the database
    @exception ItemException if there's a problem with the referenced item
    @exception CaldataException if there's a problem with the data
   */
  protected void perfAction() 
      throws SQLException, ItemException, CaldataException
  {
    if (getHandlesFunction().startsWith(CalParameter.ADD)) {
      this.cache.add((CalendarObject) this.c);
    } else if (getHandlesFunction().startsWith(CalParameter.REPLACE)) {
      cache.replace((CalendarObject) this.c);
    } else {
      throw new IllegalStateException("perfAction: illegalAction " + 
        getHandlesFunction());
    }
  }

  /**
    Perform the action on the object 

    @exception SQLException If there's a problem with the database
    @exception ItemException if there's a problem with the referenced item
    @exception CaldataException if the form data is malformed
   */
  protected void basePerformAction() 
    throws SQLException, ItemException, CaldataException
  {
    if (this.user.isGuest()) {
      return;        // guests can't do anything
    }
    
    w = null;

    // set variables we need to perform the action
    setVars();

    // perform action on the entity
    perfAction();

    message = DisplayText.addMessage(this.successMessage, message);
    
    if (w != null) {
      throw w;
    }
  }

  /**
    Print the page to the servlet's output stream
    @param resp The servlet's response object
    @exception IOException if there's a problem printing the page
    @exception java.sql.SQLException if there's a problem with the database
    @exception CaldataException if the database contains bad data
   */
  protected void baseGeneratePage(HttpServletResponse resp) 
    throws IOException, java.sql.SQLException, CaldataException
  {
    try {
      Page.create(cp.getFunction2(), cal, this.user,
                  es, message, cp).print(resp);
    } catch (ItemException e) {
      throw new RuntimeException(
        "creating simple view shouldn't cause ItemException, but it did");
    }
  }
}
