package uwcal.out;

import java.util.SortedSet;

import org.apache.ecs.Element;

import edu.washington.cac.calendar.CalParameter;
import edu.washington.cac.calendar.CalProperties;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.filter.Calendar;
import edu.washington.cac.calendar.filter.Filter;
import edu.washington.cac.calendar.filter.KeywordFilter;
import edu.washington.cac.calendar.filter.NamedFilter;
import edu.washington.cac.myuw.util.Coder;

/**
  Object holds a sorted set of calendars, plus helpers to convert
  the set into a displayable list of calendars, including, if appropriate,
  links that when selected change the current public
  events filter to the calendar in question

  @author Greg Barnes
  @version 2.1
 */
public class CalendarSet
{
  /** The current date */
  private MyCalendar mc;

  /** Page this list is a part of */
  private Page page;

  /** The set of calendars */
  private SortedSet calendars;

  /** Create a new instance 
    @param mc The current date
    @param page Page this list is a part of
   */
  public CalendarSet(MyCalendar mc, Page page, SortedSet calendars)
  {
    this.mc = mc;
    this.page = page;
    this.calendars = calendars;
  }

  /**
    Get the url to select a single calendar
    @param id Unique id of the calendar
    @return the url to select a single calendar
   */
  private String calendarUrl(int id)
  {
    return WebResources.BASE_GET_URL + CalParameter.SELECTFILTER_FUNC
        + Util.DATE_PARAM + this.mc.getDateDigits() 
        + Util.FUNCTION2_PARAM + this.page.nextFunction()
        + Util.CALENDAR_PARAM + id;
  }

  /**
    Get a single calendar name

    @param id Unique id of the calendar
    @param text Text the user sees
    @return a name for a single calendar
   */
  public Element calendarName(int id, String text)
  {
    /* guests get a link to the current view with the filter applied.
       Users just get the name
     */
    if (CalProperties.defaultIsGuest()) {
      return Util.linkItem(calendarUrl(id), text, Style.STD_TEXT);
    } else {
      return Util.spanItem(text, Style.STD_TEXT);
    }
  }

  /**
    Get the name corresponding to a Calendar
    @param c The calendar
    @return the name corresponding to the calendar
   */
  public Element name(Calendar c) 
  {
    return calendarName(c.getId(), Coder.HTMLencode(c.getFilter().getName()));
  }

  /**
    Get the url to subscribe or unsubscribe to single calendar
    @param id Unique id of the calendar
    @param subscribed If true, generate the unsubscribe URL.  Otherwise,
                      generate the subscribe URL
    @return the url to subscribe or unsubscribe to a single calendar
   */
  private String subscribeUrl(int id, boolean subscribed)
  {
    return WebResources.BASE_GET_URL 
        + (subscribed ? CalParameter.UNSUBSCRIBE_FUNC : 
                        CalParameter.SUBSCRIBE_FUNC)
        + Util.DATE_PARAM + this.mc.getDateDigits() 
        + Util.FUNCTION2_PARAM + this.page.nextFunction()
        + Util.IDENTIFIER_PARAM + this.page.getSessionIdentifier() 
        + Util.CALENDAR_PARAM + id;
  }

  /**
    Get the subscribe link corresponding to a Calendar
    @param c The calendar
    @param subscribed Is the user subscribed to the calendar?
    @return the name corresponding to the calendar
   */
  public Element subscribe(Calendar c, boolean subscribed) 
  {
    return Util.linkItem(subscribeUrl(c.getId(), subscribed), 
                         subscribed ? DisplayText.UNSUBSCRIBE :
                                      DisplayText.SUBSCRIBE, 
                         Style.AUX);
  }

  /**
    Get the set of calendars
    @return the set of calendars
   */
  public SortedSet getCalendars()
  {
    return this.calendars;
  }

  /**
    Get the set of calendars, as an array
    @return the set of calendars, as an array
   */
  public Calendar[] getCalendarArray()
  {
    return (Calendar []) this.calendars.toArray(new Calendar[0]);
  }

}
