// java

package uwcal.out;

import java.util.Iterator;

import edu.washington.cac.calendar.CalParameter;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.Keyword;
import edu.washington.cac.calendar.data.NoSuchItemException;
import edu.washington.cac.calendar.data.PersonalEvents;
import edu.washington.cac.calendar.data.PublicEvents;
import edu.washington.cac.calendar.db.DBResources;

/** A helper printer for an Event
    @author Greg Barnes
    @version 1.1
  */
public class EventHelperPrinter extends HelperPrinter 
{
  /** The event to print */
  private Event e;

  /** prefix to add to URLs */
  private String prefix = null;

  /** 
    Create a new instance 
    @param e Event to print
    @param page Page this printer is a part of
   */
  public EventHelperPrinter(Event e, Page page) 
  {
    this(e, page, null);
  }

  /** 
    Create a new instance 
    @param e Event to print
    @param page Page this printer is a part of
    @param prefix Prefix to use for urls instead of the default
   */
  public EventHelperPrinter(Event e, Page page, String prefix) 
  {
    super(e, page);
    this.e = e;
    this.prefix = prefix;
  }

  /**
    Get the edit URL for an Event
    @param today Current day
    @param sid Session identifier
   */
  String editUrl(MyCalendar today, int sid) 
  {
    return EditEvent.url(e, today, this.page, sid);
  }
 
  private static final String DELETE_EVENT_URL =
      WebResources.BASE_GET_URL + CalParameter.DELETEEVENT_FUNC + 
      Util.DATE_PARAM;

  /**
    Get the delete URL for an Event
    @param today Current day
    @param sid Session identifier
   */
  String deleteUrl(MyCalendar today, int sid) 
  {
    return Util.makeUrl(DELETE_EVENT_URL, today, this.page,
                        Util.EVENT_ID_PARAM, e.getId(), sid);
  }
  
  /** Start of the URL that removes a public event from a personal calendar */
  private static final String REMOVE_EVENT_URL =
      WebResources.BASE_GET_URL + CalParameter.REMOVEEVENTREF_FUNC + 
      Util.DATE_PARAM;

  /** Start of the URL that removes a public event with multiple instances
      from a personal calendar */
  private static final String REMOVEALL_EVENT_URL =
      WebResources.BASE_GET_URL + CalParameter.REMOVEALLEVENTREF_FUNC + 
      Util.DATE_PARAM;

  /**
    Get the eventid to use for a remove or remove all function
    @return the eventid to use for a remove or remove all function
   */
  private int eventidToRemove()
  {
    return (!e.isRecurring() || e.getRecurrence().isMaster()) ? 
           e.getId() : e.getRecurrence().getMasterId();
  }

  /**
    Get the URL to remove the event
    @param today the current day
    @param sid Session id
    @return the URL to remove the event
   */
  String removeUrl(MyCalendar today, int sid) 
  {
    return (hasMultipleInstances() ? REMOVEALL_EVENT_URL : REMOVE_EVENT_URL) + 
          today.getDateDigits() + Util.FUNCTION2_PARAM + 
          this.page.nextFunction() +
          Util.EVENT_ID_PARAM + eventidToRemove() +
          Util.IDENTIFIER_PARAM + sid;
  }
  
  String cName() 
  {
    return DisplayText.EVENT;
  }

  /** 
    Can this event be removed (c.f. The remove function)? 
    @return Can the event be removed (c.f. The remove function)? 
   */
  boolean canBeRemoved()
  {
    return !this.page.getUser().isGuest() && 
           ((PersonalEvents) this.page.getEvents()).explicitlyContains(this.e);
  }

  /**
    Get the URL to remove all instances of the event
    @param today the current day
    @param sid Session id
    @return the URL to remove the event
   */
  String removeAllUrl(MyCalendar today, int sid) 
  {
    return removeUrl(today, sid);
  }
  
  /** 
    Is the event one instance of many
    @return Is the event on instance of many
   */
  boolean hasMultipleInstances()
  {
    return e.isRecurring();
  }

  /** Is the event official
    @return Is an event 'official'? 
   */
  boolean isOfficial()
  {
    Iterator keywords = e.getKeywords();
    
    while (keywords.hasNext()) {
      Integer id = (Integer) keywords.next();
      Keyword k;

      try {
        k = PublicEvents.getPublicKeywords().getKeyword(id.intValue());
      } catch (NoSuchItemException ex) {
        throw new RuntimeException("Non-existent keyword " + id + " for event "
                                   + e.getId());
      }

      if (k.getCreator().equals(DBResources.officialKeywordCreator())) {
        return true;
      }
    }

    return false;
  }

  /**
    Get the URL for the MORE link
    @param today The current date
    @return the URL for the MORE link
   */
  String moreUrl(MyCalendar today) 
  {
    return Eventview.url(this.e, today, this.page, prefix);
  }
}
