// java

package uwcal.out;

import java.text.DateFormat;

import org.apache.ecs.Element;
import org.apache.ecs.ElementContainer;
import org.apache.ecs.Entities;
import org.apache.ecs.html.Span;

import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.Event;
import edu.washington.cac.calendar.data.EventsI;
import edu.washington.cac.calendar.data.Location;
import edu.washington.cac.calendar.data.User;
import edu.washington.cac.myuw.util.Coder;

/** 
  An object that prints an event
  @author Greg Barnes
  @version 1.1
 */
public abstract class EventPrinter 
{
  /** The user for whom the event is being printed */
  protected User user;

  /** The current day */
  protected MyCalendar today;

  /** The view this printer is a part of */
  protected Timeview view;

  /**
    Create a new instance
    @param user The user for whom the event will be printed
    @param today The current day
    @param view View the printer is a part of
   */
  protected EventPrinter(User user, 
                         MyCalendar today, 
                         Timeview view) 
  { 
    this.user = user;
    this.today = today;
    this.view = view;
  }

  /**
    Set the current date being printed
    @param today The current date being printed
   */
  void setDate(MyCalendar today) 
  {
    this.today = today;
  }

  /**
    Get the current date being printed
    @return The current date being printed
   */
  MyCalendar getDate() 
  {
    return this.today;
  }

  protected abstract String linkStyle();
  protected abstract String spanStyle();
  protected abstract Element format(Event e, EventsI es);
  protected abstract String noTimeRange();

  /**
    Get the first day represented by the surrounding page
    @return the first day represented by the surrounding page
   */
  MyCalendar getFirstDay()
  {
    return this.view.getFirstDay();
  }

  /**
    Get the last day represented by the surrounding page
    @return the last day represented by the surrounding page
   */
  MyCalendar getLastDay()
  {
    return this.view.getLastDay();
  }

  protected String timeStyle() 
  {
    return Style.TIME;
  }

  /**
    Does the date/time object contain any time information?
    @param mc date/time object
    @return Does the date/time object contain any time information?
   */
  protected boolean hasTime(MyCalendar mc) 
  {
    return mc != null && mc.hasTime();
  }

  /**
    Get the time portion of a date/time
    @param mc date/time
    @return the time portion of a date/time
   */
  protected String timeString(MyCalendar mc) 
  {
    return hasTime(mc) ? mc.timeString().toLowerCase() : "";
  }

  /**
    Is this day different from another
    @param mc The other day
    @return Is this day different from another
   */
  private boolean differentDay(MyCalendar mc) 
  {
    return !mc.isSameDay(today);
  }

  /**
    Is the current day the interesting boundary day for a start/end date?
    @param start If true we are interested in start dates.  
                 If false, we are interested in end dates.
    @return If start is true, whether we are printing the first day of the view
            If start is false, whether we are printing the last day of the view
   */
  private boolean currentDayIsBoundaryDay(boolean start) 
  {
    return start ? this.view.isFirstDay(this.today) : 
                   this.view.isLastDay(this.today);
  }

  /**
    Is this date interesting enough to print?
    @param mc The date
    @param start Is the date a start or end date?
    @return Is this date interesting enough to print?
   */
  protected boolean shouldPrintDate(MyCalendar mc, boolean start)
  {
    return differentDay(mc) && 
           (mc.hasTime() || currentDayIsBoundaryDay(start));
  }

  /**
    What to put between a date and a time
    @return What to put between a date and a time
   */
  protected String dateTimeSeparator()
  {
    return " ";
  }
  
  /**
    Get a date/time as HTML
    @param mc The date/time
    @param printDate Should the date be printed?
    @param dateFormat The <code>java.text.DateFormat</code> for the date 
   */
  protected Element dateTime(MyCalendar mc, 
                             boolean printDate,
                             int dateFormat) 
  {
    Span s = new Span();
    s.setClass(Style.PRE);

    if (mc == null) {
      return new ElementContainer();
    } 

    if (printDate) {
      s.addElement(mc.dateString(dateFormat));
    
      if (hasTime(mc)) {
        s.addElement(dateTimeSeparator());
      }
    }
    
    if (hasTime(mc)) {
      s.addElement(timeString(mc));
    }

    return s;
  }

  /**
    Get a date/time as HTML
    @param mc The date/time
    @param printDate Should the date be printed?
   */
  protected Element dateTime(MyCalendar mc, 
                             boolean printDate) 
  {
    return dateTime(mc, printDate, DateFormat.SHORT);
  }

  /**
    Will any part of a date/time be printed?
    @param mc The date/time
    @param start Does the date/time represent a start date (or end date)
   */
  private boolean nothingToPrint(MyCalendar mc, boolean start) 
  {
    return mc == null || 
        (!mc.hasTime() && !shouldPrintDate(mc, start));
  }

  /**
    Does an event have no date/time information to print?
    @param e The event
    @return Does an event have no date/time information to print?
   */
  protected boolean hasNoTime(Event e) 
  {
    return nothingToPrint(e.startCalendar(), true) &&
           nothingToPrint(e.endCalendar(), false);
  }

  /**
    Assuming we're printing something for the date/time,
    should we see a hyphen (between the start and end dates)?

    @param e Event we're printing
    @return Should we see a hyphen (between the start and end dates)?
   */
  protected boolean shouldPrintHyphen(Event e)
  {
    return e.endCalendar() != null;
  }

  /**
    Get the date/time info for the event
    @param e The event
    @return the date/time info for the event
   */
  protected Element timeRange(Event e) 
  {
    if (hasNoTime(e)) {
      return new ElementContainer().addElement(noTimeRange());
    } else {
      ElementContainer ec = new ElementContainer().
          addElement(dateTime(e.startCalendar(), 
                              shouldPrintDate(e.startCalendar(), true)));

      if (shouldPrintHyphen(e)) {
        ec.addElement(" - ");
      }

      ec.addElement(dateTime(e.endCalendar(), 
                             shouldPrintDate(e.endCalendar(), false)));

      return Util.spanItem(ec, timeStyle());
    }
  }

  /**
    Get the add icon for this event, if appropriate
    @param e The event in question
    @return The icon, as HTML
   */
  protected Element addIcon(Event e)
  {
    if (this.user.isGuest()) {
      return new ElementContainer(new AddEventRef().icon(e)).
                 addElement(Entities.NBSP);
    } else {
      return new ElementContainer();
    }
  }

  /**
    Get the short description of an event
    @param e The event
    @return the short description of an event, as HTML
   */
  protected Element eventDesc(Event e) 
  {
    if (e.getLink() != null) {
      return Util.linkItem(e.getLink(), Coder.HTMLencode(e.getShortdesc()), 
                           linkStyle());
    } else {
      return Util.spanItem(Coder.HTMLencode(e.getShortdesc()), spanStyle());
    }
  }

    private static final String LOCATION_PREFIX = ", ";

  protected Element location(Event e, EventsI es) 
  {
    Location l = e.getLocation(es);

    if (!l.isLocation()) {
      return new ElementContainer();
    } else if (l.getLink() != null) {
      return new ElementContainer().
          addElement(LOCATION_PREFIX).
          addElement(Util.linkItem(l.getLink(), 
                                   Coder.HTMLencode(l.getAddress()), 
                                   linkStyle()));
    } else {
      return Util.spanItem(LOCATION_PREFIX + Coder.HTMLencode(l.getAddress()), 
                           spanStyle());
    }
  }
}
