// java

package uwcal.out.forms;

import java.util.SortedSet;

import org.apache.ecs.AlignType;
import org.apache.ecs.Element;
import org.apache.ecs.ElementContainer;
import org.apache.ecs.Entities;
import org.apache.ecs.html.BR;
import org.apache.ecs.html.Input;
import org.apache.ecs.html.Select;
import org.apache.ecs.html.Option;
import org.apache.ecs.html.Table;
import org.apache.ecs.html.TD;

import edu.washington.cac.calendar.CalParameter;
import edu.washington.cac.calendar.MyCalendar;
import edu.washington.cac.calendar.data.CaldataException;
import edu.washington.cac.calendar.data.Keyword;
import edu.washington.cac.calendar.data.Keywords;
import edu.washington.cac.calendar.data.PublicEvents;
import edu.washington.cac.calendar.data.Sponsor;
import edu.washington.cac.calendar.data.Sponsors;
import edu.washington.cac.calendar.db.DBResources;
import edu.washington.cac.calendar.filter.Calendar;
import edu.washington.cac.calendar.filter.Filter;
import edu.washington.cac.calendar.filter.KeywordFilter;
import edu.washington.cac.calendar.filter.NamedFilter;
import edu.washington.cac.calendar.filter.OrFilter;
import edu.washington.cac.calendar.filter.PublicCalendars;
import edu.washington.cac.calendar.filter.SearchFilter;
import edu.washington.cac.calendar.filter.SelectAllFilter;
import edu.washington.cac.calendar.filter.SponsorFilter;
import edu.washington.cac.myuw.util.Coder;

import uwcal.CalProfile;
import uwcal.out.AllCalendars;
import uwcal.out.CalendarSet;
import uwcal.out.Page;
import uwcal.out.Style;
import uwcal.out.Util;
import uwcal.out.WebResources;

/**
  A form for a guest user to select the current filter on the public events
  @author Greg Barnes
  @version 1.2
 */
public class SelectFilterForm extends CalendarForm
{
  /** The currently selected filter*/
  private Filter filter;

  /** Helper class to create form elements */
  private CommonFormElements cfe = new CommonFormElements();

  /** The set of calendars to appear on the form */
  private CalendarSet mcl;

  /**
    Create a new instance
    @param filter The currently selected filter
    @param page Page the form is a part of
   */
  public SelectFilterForm(Page page, Filter filter)
  {
    super(page);

    if (filter instanceof SelectAllFilter) {
      // use a rational default instead 
      this.filter = CalProfile.DEFAULT_FILTER;
    } else {
      this.filter = filter;
    }

    this.mcl = new CalendarSet(page.getMc(), page, 
                               PublicCalendars.mainCalendars());
  }

  /** The prefix before the name of a special filter in the selection list*/
  private static final String SPECIAL_FILTER = "Special";

  /**
    Get the function that the form triggers
    @return the function that the form triggers
   */
  protected String formFunction() 
  {
    return CalParameter.SELECTFILTER_FUNC;
  }

  /**
    Get the title of the form
    @return the title of the form
   */
  protected String titleText() 
  {
    return FormsDisplayText.SELECT_FILTER_TITLE;
  }

  /**
    Get the text before the search box
    @return the text before the search box
   */
  private Element searchName()
  {
    return Util.spanItem(FormsDisplayText.SEARCH, Style.STD_TEXT);
  }

  /**
    Get the text inside search box
    @return the text inside the search box
   */
  private String searchText()
  {
    return this.filter instanceof SearchFilter ? 
        Coder.HTMLencode(this.filter.getName()) : "";
  }

  /**
    Get the search box
    @return the search box
   */
  private Element searchBox()
  {
    return new Input(Input.TEXT, CalParameter.SEARCH_STRING, searchText()).
           setSize(30);
  }

  /**
    Get the search subform as HTML
    @return the search subform as HTML
   */
  private Element search()
  {
    return cfe.tableRowOneCell(new ElementContainer().
                               addElement(searchName()).
                               addElement(Entities.NBSP).
                               addElement("\n").
                               addElement(searchBox()).
                               addElement("\n").
                               addElement(GoForm.goButton()).
                               addElement("\n"));
  }

  /** The most calendars we will display in single column mode */
  private static final int MAX_SINGLE_COLUMN = 0;

  /**
    Get the list of main calendars as HTML
    @return the list of main calendars as HTML
   */
  private Element calendarList()
  {
    ElementContainer ec = new ElementContainer();
    Calendar[] calendars = mcl.getCalendarArray();

    if (calendars.length > MAX_SINGLE_COLUMN) {
      /* Want to split into two columns.  There's some trickery involved
         here if there are an odd number of elements.  We want the
         left column to get the extra element.
       */
      for (int i = 0; i < calendars.length/2; i++) {
        ec.addElement(cfe.tableRowTwoCells(
            mcl.name(calendars[i]),
            mcl.name(calendars[i + (calendars.length + 1)/2])));
        // calendars.size() + 1 ensures the left column gets the extra if there
        // are an odd number of boxes
      }

      // If calendars.size() is odd, add the extra element
      if (calendars.length/2 != (calendars.length + 1)/2) {     
        ec.addElement(cfe.tableRowTwoCells(
            mcl.name(calendars[calendars.length/2]),
            new ElementContainer()));
      }
    } else {
      for (int i = 0; i < calendars.length; i++) {
        ec.addElement(cfe.tableRowOneCell(new TD().setColSpan(2).
            addElement(mcl.name(calendars[i]))));
      }
    }

    return ec;
  }

  /**
    Get the row that selects the default calendar
    @return the row that selects the default calendar
   */
  private Element defaultCalendar()
  {
    return cfe.tableRowOneCell(
        new TD().setColSpan(2).
            addElement(mcl.name(PublicCalendars.defaultCalendar())));
  }
    
  /** 
     Get the title cell above the main calendar
     @return the title cell above the main calendar
   */
  private Element mainCalendarTitle()
  {
    return new ElementContainer().
           addElement(cfe.headerRow(FormsDisplayText.MAIN_CALENDARS_TITLE,
                                    Style.LOCATION_HEADER)).
           addElement("\n");
  }

  /**
    Get the list of main categories as HTML
    @return the list of main categories as HTML
   */
  private Element mainCalendars()
  {
    return new ElementContainer().
           addElement(cfe.tableRowOneCell(
               new TD().setColSpan(2).addElement(Util.spanItem(
                   FormsDisplayText.MAIN_HEADER, Style.STD_TEXT)))).
           addElement("\n").addElement(cfe.blankRow()).
           addElement(calendarList()).
           addElement("\n").addElement(cfe.blankRow()).
           addElement(defaultCalendar()).
           addElement("\n");
  }

  /**
    Get the list of other categories as HTML
    @return the list of other categories as HTML
   */
  private Element otherCalendars()
  {
    return cfe.tableRowOneCell(
        new TD().setColSpan(2).setAlign(AlignType.CENTER).
            addElement(Util.linkItem(AllCalendars.url(getMc(), this.page),
                                     FormsDisplayText.ALL_CALENDARS_TEXT,
                                     Style.STD_TEXT)));
  }

  /**
    Get the list of calendars 
    @return the list of calendars 
   */
  protected Element calendars()
  {
    return new ElementContainer().
           addElement(
               cfe.tableRowOneCell(new TD().setColSpan(2).
                                   addElement(new Table().
                                              addElement(mainCalendarTitle()).
                                              addElement(mainCalendars()).
                                              addElement(cfe.blankRow()).
                                              addElement(otherCalendars()).
                                              setClass(Style.FORM_ELEMENT))));
  }

  /**
    Get the elements in the form
    @return the elements in the form
   */
  protected Element[] inputElements() 
  {
    Element[] e = { search(), 
                    cfe.blankRow(Style.SPACER_HEADER), calendars() };
    return e;
  }

  /**
    Get an array that says whether to add borders to the individual 
    element tables
    @return an array that says whether to add borders to the individual 
    element tables, null if there should be no such borders
   */
  protected boolean[] setElementBorders() 
  {
    boolean[] b = { true, false, true };
    return b;
  }
    
  /**
    Get the text shown in the submit button
    @return the text shown in the submit button
   */
  protected String submitButtonText()
  { 
    return FormsDisplayText.FILTER;
  }

  /**
    Get the end of the form
    @return The end of the form
   */
  protected Element end() 
  {
    return new ElementContainer().
           addElement(cfe.blankRow(Style.FORM_TITLE));
  }
}
